#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>

#include "queue.h"

void init_queue(queue_t *q)
{
  memset(q, 0, sizeof(*q));
  q->mutex = malloc(sizeof(pthread_mutex_t)); if (q->mutex==NULL) abort();
  q->cond = malloc(sizeof(pthread_cond_t)); if (q->cond==NULL) abort();
  if (pthread_mutex_init(q->mutex, NULL)) abort();
  if (pthread_cond_init(q->cond, NULL)) abort();
  q->alive = 1;
}

void free_queue(queue_t *q)
{
  free(q->mutex);
  free(q->cond);
}

void enqueue(queue_t *q, char *buf)
{
  if (pthread_mutex_lock(q->mutex)) abort();
  if (!q->alive) goto end;
  if (q->size == QUEUE_SIZE)
    { printf("queue full, drop packet\n"); goto end; }

  memcpy(q->e[q->write], buf, 4096);
  q->write++; q->write %= QUEUE_SIZE;
  q->size++;

end:
  if (pthread_cond_signal(q->cond)) abort();
  if (pthread_mutex_unlock(q->mutex)) abort();
#if 0
  if (pthread_mutex_lock(q->mutex)) abort();
  while (q->alive && q->size == QUEUE_SIZE)
    if (pthread_cond_wait(q->cond, q->mutex)) abort();
  if (!q->alive) goto end;

  memcpy(q->e[q->write], buf, 4096);
  q->write++; q->write %= QUEUE_SIZE;
  q->size++;

end:
  if (pthread_cond_signal(q->cond)) abort();
  if (pthread_mutex_unlock(q->mutex)) abort();
#endif
}

void dequeue(queue_t *q, char *buf)
{
  if (pthread_mutex_lock(q->mutex)) abort();
  while (q->alive && q->size == 0)
    if (pthread_cond_wait(q->cond, q->mutex)) abort();
  if (!q->alive) goto end;

  memcpy(buf, q->e[q->read], 4096);
  q->read++; q->read %= QUEUE_SIZE;
  q->size--;

end:
  if (pthread_cond_signal(q->cond)) abort();
  if (pthread_mutex_unlock(q->mutex)) abort();
}

void kill_queue(queue_t *q)
{
  if (pthread_mutex_lock(q->mutex)) abort();

  q->alive = 0;

  if (pthread_cond_broadcast(q->cond)) abort();
  if (pthread_mutex_unlock(q->mutex)) abort();
}
