#ifndef _UTILS_H_
#define _UTILS_H_

#include <stdint.h>
#include <sys/time.h>

/*
description of streams working:
http://0pointer.de/lennart/projects/pulseaudio/doxygen/streams.html
*/

#if 0
    PA_SAMPLE_U8,          /**< Unsigned 8 Bit PCM */
    PA_SAMPLE_ALAW,        /**< 8 Bit a-Law */
    PA_SAMPLE_ULAW,        /**< 8 Bit mu-Law */
    PA_SAMPLE_S16LE,       /**< Signed 16 Bit PCM, little endian (PC) */
    PA_SAMPLE_S16BE,       /**< Signed 16 Bit PCM, big endian */
    PA_SAMPLE_FLOAT32LE,   /**< 32 Bit IEEE floating point, little endian (PC), range -1.0 to 1.0 */
    PA_SAMPLE_FLOAT32BE,   /**< 32 Bit IEEE floating point, big endian, range -1.0 to 1.0 */
    PA_SAMPLE_S32LE,       /**< Signed 32 Bit PCM, little endian (PC) */
    PA_SAMPLE_S32BE,       /**< Signed 32 Bit PCM, big endian */
    PA_SAMPLE_S24LE,       /**< Signed 24 Bit PCM packed, little endian (PC). \since 0.9.15 */
    PA_SAMPLE_S24BE,       /**< Signed 24 Bit PCM packed, big endian. \since 0.9.15 */
    PA_SAMPLE_S24_32LE,    /**< Signed 24 Bit PCM in LSB of 32 Bit words, little endian (PC). \since 0.9.15 */
    PA_SAMPLE_S24_32BE,    /**< Signed 24 Bit PCM in LSB of 32 Bit words, big endian. \since 0.9.15 */
#endif

#define IN_MAX 65536

#define CHANNEL_MAX 32

typedef struct {
  int format;
  int channels;
  uint32_t rate;
} sample_spec_t;

typedef struct {
  int count;
  int map[CHANNEL_MAX];
} channel_map_t;

typedef struct {
  int count;
  uint32_t volume[CHANNEL_MAX];
} cvolume_t;

typedef enum {
  PA_COMMAND_REPLY                   = 2,
  PA_COMMAND_CREATE_PLAYBACK_STREAM  = 3,
  PA_COMMAND_AUTH                    = 8,
  PA_COMMAND_SET_CLIENT_NAME         = 9,
  PA_COMMAND_GET_SINK_INFO_LIST      = 22,
  PA_COMMAND_REQUEST                 = 61,
  PA_COMMAND_DRAIN_PLAYBACK_STREAM   = 12,
  PA_COMMAND_DELETE_PLAYBACK_STREAM  = 4,
  PA_COMMAND_GET_SOURCE_INFO_LIST    = 24,
  PA_COMMAND_SUBSCRIBE               = 35,
  PA_COMMAND_GET_PLAYBACK_LATENCY    = 14,
  PA_COMMAND_GET_SERVER_INFO         = 20,
  PA_COMMAND_GET_SINK_INFO           = 21,
  PA_COMMAND_SET_SINK_INPUT_VOLUME   = 37,
  PA_COMMAND_CORK_PLAYBACK_STREAM    = 41,
  PA_COMMAND_TRIGGER_PLAYBACK_STREAM = 43,
  PA_COMMAND_GET_SOURCE_INFO         = 23,
  PA_COMMAND_CREATE_RECORD_STREAM    = 5,
  PA_COMMAND_FLUSH_PLAYBACK_STREAM   = 42,
  PA_COMMAND_GET_RECORD_LATENCY      = 57,
  PA_COMMAND_CORK_RECORD_STREAM      = 58,
  PA_COMMAND_FLUSH_RECORD_STREAM     = 59,
  PA_COMMAND_DELETE_RECORD_STREAM    = 6,
  VOLUME_CLIENT                      = 666,
} command_type;

typedef struct {
  uint32_t protocol_version;
  char cookie[IN_MAX];
  int cookie_size;
} command_auth_t;

typedef struct {
  uint32_t protocol;
} command_reply_t;

typedef struct {
  uint32_t index;
  uint32_t length;
} command_request_t;

typedef struct {
  uint32_t index;
} command_drain_t;

typedef struct {
  uint32_t index;
} command_delete_playback_t;

typedef struct {
  int mask;
} command_subscribe_t;

typedef struct {
  char media_name[IN_MAX];
  sample_spec_t spec;
  channel_map_t map;
  char dev[IN_MAX];
  uint32_t maxlength;
  int corked;
  uint32_t tlength;
  uint32_t prebuf;
  uint32_t minreq;
  uint32_t syncid;
  cvolume_t volume;
} command_playback_t;

typedef struct {
  uint32_t idx;
  struct timeval time;
} command_playback_latency_t;

typedef struct {
  int idx;
  char name[IN_MAX];
} command_sink_info_t;

typedef struct {
  int idx;
  cvolume_t volume;
} command_set_sink_input_volume_t;

typedef struct {
  int idx;
  int bool;
} command_cork_playback_t;

typedef struct {
  int idx;
} command_trigger_t;

typedef struct {
  int idx;
  char name[IN_MAX];
} command_source_info_t;

typedef struct {
  char name[IN_MAX];
  sample_spec_t spec;
  channel_map_t map;
  int source_index;
  char source_name[IN_MAX];
  uint32_t maxlength;
  int corked;
  uint32_t fragsize;
} command_record_t;

typedef struct {
  int idx;
} command_flush_playback_t;

typedef struct {
  uint32_t idx;
  struct timeval time;
} command_record_latency_t;

typedef struct {
  int idx;
  int bool;
} command_cork_record_t;

typedef struct {
  int idx;
} command_flush_record_t;

typedef struct {
  uint32_t index;
} command_delete_record_t;

typedef struct {
  uint32_t length;
  uint32_t channel;
  uint32_t offset_hi;
  uint32_t offset_lo;
  uint32_t flags;
  command_type type;
  uint32_t tag;
  union {
    command_auth_t auth;
    command_reply_t reply;
    command_playback_t playback;
    command_request_t request;
    command_drain_t drain;
    command_delete_playback_t delete_playback;
    command_subscribe_t subscribe;
    command_playback_latency_t playback_latency;
    command_sink_info_t sink_info;
    command_set_sink_input_volume_t si_volume;
    command_cork_playback_t cork_playback;
    command_trigger_t trigger;
    command_source_info_t source_info;
    command_record_t record;
    command_flush_playback_t flush_playback;
    command_record_latency_t record_latency;
    command_cork_record_t cork_record;
    command_flush_record_t flush_record;
    command_delete_record_t delete_record;
  } c;
} message;

typedef struct {
  int len;
  unsigned char buf[IN_MAX];
  int pos;
  void *x;         /* to store some stuff internally if needed */
} buffer;

void read_message(int fd, buffer *r, message *m);

void new_thread(void *(*f)(void *), void *data);

void putu32(buffer *b, uint32_t v);
void puts64(buffer *b, int64_t v);
void put_bool(buffer *b, int v);
void put_usec(buffer *b, uint64_t v);
void put_timeval(buffer *b, struct timeval v);
void put_string(buffer *b, char *s);
void put_sample_spec(buffer *b, int format, int channels, int samplerate);
void put_channel_map(buffer *b, int n, int *channels);
void put_cvolume(buffer *b, int n, int *volumes);

struct timeval get_time_now(void);

void new_reply(buffer *b, int channel, int tag);
void size_reply(buffer *b);
void new_reply_more_data(buffer *b, int channel, int size);

int fullwrite(int fd, void *_buf, int count);
int fullread(int fd, void *_buf, int count);

#endif /* _UTILS_H_ */
