#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <X11/Xlib.h>
#include <X11/Xft/Xft.h>
#include <poll.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <unistd.h>
#include <time.h>
#include <stdint.h>

/**
 *  \brief R3(x, A, B, C, D):
 *            f(x) calculated by linear interpolation
 *            where f(A) = C and f(B) = D.
 */
#define R3(x, A, B, C, D) \
    ((double)(C) + ((double)(x)-(double)(A)) * ((double)(D)-(double)(C)) \
                                             / ((double)(B)-(double)(A)))

#define R(x)if(x)do{printf("%s:%d: err\n",__FILE__,__LINE__);exit(1);}while(0)

typedef struct { int id; int vol; } vol;

#define BG 0
#define FG 1
#define ACTIVE 2
#define N_COLORS 3

typedef struct {
  Display *d;
  Window w;
  Pixmap p;
  int width, height;   /* width/height of window */
  int x, y;            /* position of mouse */
  int scroll_start;    /* horizontal scroll start position */
  GC colors[N_COLORS];
  XftColor xft_colors[N_COLORS];
  int text_height;
  int baseline;
  XftFont *font;
  XftDraw *xft;
  int scrollbar_height;
  int unit_width;
  int scrollbar_emphasized;   /* 0: no, 1: yes */
  int vol_emphasized;         /* -1: none, n>=0: this one */
} gui;

void compute_emphasized(gui *g)
{
  g->scrollbar_emphasized = g->y >= g->height - g->scrollbar_height;
  g->vol_emphasized = -1;
  if (g->x >= 0 && g->y < g->height - g->scrollbar_height)
    g->vol_emphasized = (g->x + g->scroll_start) / g->unit_width;
}

int get_text_width(gui *g, char *t)
{
  XGlyphInfo ext;
  XftTextExtentsUtf8(g->d, g->font, (FcChar8 *)t, strlen(t), &ext);
  return ext.width;
}

void draw(gui *g, vol *v, int n)
{
  int y_start = g->text_height + 2 + 2;
  int y_end = g->height - (g->text_height + 2 + 2) - g->scrollbar_height;
  int i;
  int x;
  int y;
  char t[64];
  int text_width;
  int x_scroll_start, x_scroll_end;
  compute_emphasized(g);
  XFillRectangle(g->d, g->p, g->colors[BG], 0, 0, g->width, g->height);
  if (n == 0) {
    XftDrawStringUtf8(g->xft, &g->xft_colors[FG], g->font,
                      2, 2+g->baseline, (unsigned char *)"no stream", 9);
    return;
  }
  for (i = 0; i < n; i++) {
    int vv;
    x = g->unit_width * i - g->scroll_start;
    if (i == g->vol_emphasized)
      XFillRectangle(g->d, g->p, g->colors[ACTIVE],
                     x+2, y_start - 2,
                     g->unit_width-4 + 1, y_end - y_start + 4 + 1);
    sprintf(t, "%d", v[i].id);
    text_width = get_text_width(g, t);
    XftDrawStringUtf8(g->xft, &g->xft_colors[FG], g->font,
                      x + (g->unit_width - text_width) / 2, 2+g->baseline,
                      (unsigned char *)t, strlen(t));
    if (v[i].vol > -100) sprintf(t, "%d", v[i].vol);
    else                 sprintf(t, "[%d]",v[i].vol + 1000);
    text_width = get_text_width(g, t);
    XftDrawStringUtf8(g->xft, &g->xft_colors[FG], g->font,
            x + (g->unit_width - text_width) / 2, y_end + 2 + 2+g->baseline,
            (unsigned char *)t, strlen(t));
    XDrawLine(g->d, g->p, g->colors[FG],
              x+g->unit_width/2, y_start,
              x+g->unit_width/2, y_end);
    vv = v[i].vol; if (vv < -100) vv += 1000;
    y = R3(vv, -60, 60, y_end, y_start);
    XDrawLine(g->d, g->p, g->colors[FG],
              x+g->unit_width/2-5, y,
              x+g->unit_width/2+5, y);
    XDrawLine(g->d, g->p, g->colors[FG],
              x+g->unit_width/2-5, y+1,
              x+g->unit_width/2+5, y+1);
  }
  if (g->scrollbar_emphasized)
    XFillRectangle(g->d, g->p, g->colors[ACTIVE],
                   0, g->height - g->scrollbar_height,
                   g->width, g->scrollbar_height);
  x_scroll_start = R3(g->scroll_start, 0, n * g->unit_width, 0, g->width-5);
  if (x_scroll_start < 0) x_scroll_start = 0;
  if (x_scroll_start > g->width-10) x_scroll_start = g->width-10;
  x_scroll_end = R3(g->scroll_start + g->width,
                    0, n * g->unit_width, 0, g->width-5);
  if (x_scroll_end < x_scroll_start + 20) x_scroll_end = x_scroll_start + 20;
  if (x_scroll_end > g->width-5) x_scroll_end = g->width-5;
  XDrawLine(g->d, g->p, g->colors[FG],
            2+x_scroll_start+1, g->height-7, 2+x_scroll_end-1, g->height-7);
  XDrawLine(g->d, g->p, g->colors[FG],
            2+x_scroll_start, g->height-6, 2+x_scroll_end, g->height-6);
  XDrawLine(g->d, g->p, g->colors[FG],
            2+x_scroll_start+1, g->height-5, 2+x_scroll_end-1, g->height-5);
}

GC new_color(Display *d, char *color, XftColor *c)
{
  GC ret = XCreateGC(d, DefaultRootWindow(d), 0, NULL);
  XGCValues gcv;
  XColor rcol, scol;
  XCopyGC(d, DefaultGC(d, DefaultScreen(d)), -1L, ret);
  R(XAllocNamedColor(d, DefaultColormap(d, DefaultScreen(d)),
                      color, &scol, &rcol) == 0);
  gcv.foreground = scol.pixel;
  XChangeGC(d, ret, GCForeground, &gcv);
  R(XftColorAllocName(d, DefaultVisual(d, DefaultScreen(d)),
                      DefaultColormap(d, DefaultScreen(d)),
                      color, c) == False);
  return ret;
}

static void write_int(int fd, int v)
{
  char b[4];
  b[0] = v & 255;
  b[1] = (v >>  8) & 255;
  b[2] = (v >> 16) & 255;
  b[3] = (v >> 24) & 255;
  if (write(fd, b, 4) != 4) exit(1);
}

int get_int(int fd)
{
  char b[4];
  if (read(fd, b, 4) != 4) exit(1);
  return ( b[0] & 255) |
         ((b[1] & 255) <<  8) |
         ((b[2] & 255) << 16) |
         ((b[3] & 255) << 24);
}

uint64_t gettime(void)
{
  struct timespec t;
  R(clock_gettime(CLOCK_MONOTONIC_RAW, &t) == -1);
  return (uint64_t)t.tv_sec * 1000 + (uint64_t)t.tv_nsec / 1000000;
}

int vol_cmp(const void *_a, const void *_b)
{
  const vol *a = _a;
  const vol *b = _b;
  return a->id - b->id;
}

int main(void)
{
  int paint, plot, resize;
  gui g;
  struct pollfd pfd[2];
  int width = 640;
  int height = 180;
  int new_width = 0, new_height = 0;
  int volume_changed = 0;
  int fd;
  struct sockaddr_un addr;
  int n = 0;
  vol *v = NULL;
  int i;
  int timeout;
  uint64_t last_plot = gettime() - 1000;

  R((fd = socket(AF_UNIX, SOCK_STREAM, 0)) == -1);
  addr.sun_family = AF_UNIX;
  addr.sun_path[0] = 0;
  strcat(addr.sun_path, "/tmp/pulseaudionano.socket");
  R(connect(fd, (struct sockaddr *)&addr, sizeof(addr)) == -1);
  { char msg[] = { 0x12, 0x34, 0x56, 0x78,
                   0xff, 0xff, 0xff, 0xff,
                   0x34, 0x56, 0x78, 0x90,
                   0x45, 0x67, 0x89, 0x01,
                   0x56, 0x78, 0x90, 0x12, };
    R(write(fd, msg, sizeof(msg)) != sizeof(msg)); }

  g.d = XOpenDisplay(0); R(g.d==NULL);
  g.w = XCreateSimpleWindow(g.d, DefaultRootWindow(g.d), 0, 0,
                            width, height, 0, 0, 0);
  XSelectInput(g.d, g.w, KeyPressMask | ExposureMask | PointerMotionMask |
               ButtonPressMask | StructureNotifyMask |
               EnterWindowMask | LeaveWindowMask);
  XMapWindow(g.d, g.w);
  g.p = XCreatePixmap(g.d, g.w, width, height,
                      DefaultDepth(g.d, DefaultScreen(g.d)));
  g.width = width;
  g.height = height;
  g.x = g.y = -1;
  g.scroll_start = 0;
  g.colors[BG]     = new_color(g.d, "#002", &g.xft_colors[BG]);
  g.colors[FG]     = new_color(g.d, "lightblue", &g.xft_colors[FG]);
  g.colors[ACTIVE] = new_color(g.d, "#224", &g.xft_colors[ACTIVE]);
  R((g.font = XftFontOpenName(g.d, DefaultScreen(g.d), "sans-12")) == NULL);
  g.text_height = g.font->height;
  g.baseline    = g.font->ascent;
  R((g.xft = XftDrawCreate(g.d, g.p, DefaultVisual(g.d, DefaultScreen(g.d)),
      DefaultColormap(g.d, DefaultScreen(g.d)))) == NULL);
  g.scrollbar_height = 20;
  g.unit_width       = 50;

  plot = 1;

  pfd[0].fd = ConnectionNumber(g.d);
  pfd[0].events = POLLIN;
  pfd[1].fd = fd;
  pfd[1].events = POLLIN;

  while (1) {
    while (XPending(g.d)) {
      XEvent e;
x:    XNextEvent(g.d, &e);
      switch (e.type) {
      case MapNotify:
      case Expose: paint = 1; break;
      case ConfigureNotify:
        new_width = e.xconfigure.width;
        new_height = e.xconfigure.height;
        if (new_width < 10) new_width = 10;
        if (new_height < 10) new_height = 10;
        resize = 1;
        break;
      case KeyPress: {
        char buf[65];
        KeySym ks;
        compute_emphasized(&g);
        XLookupString(&e.xkey, buf, 64, &ks, NULL);
        switch (ks) {
          case XK_Escape: exit(0);
          case 'm':
            /* mute/unmute */
            if (g.vol_emphasized >= 0 && g.vol_emphasized < n) {
              if (v[g.vol_emphasized].vol < -100)
                v[g.vol_emphasized].vol += 1000;
              else
                v[g.vol_emphasized].vol -= 1000;
              volume_changed = 1;
              plot = 1;
            }
            break;
          case XK_Up:
            /* volume up (also settable when muted) */
up:         if (g.vol_emphasized >= 0 && g.vol_emphasized < n) {
              if ((v[g.vol_emphasized].vol > -100 &&
                   v[g.vol_emphasized].vol < 60) ||
                  (v[g.vol_emphasized].vol < -100 &&
                   v[g.vol_emphasized].vol < -1000+60)) {
                v[g.vol_emphasized].vol++;
                volume_changed = 1;
                plot = 1;
              }
            }
            /* scroll view */
            else if (g.scrollbar_emphasized) {
              if (g.scroll_start + g.width < n * g.unit_width) {
                g.scroll_start += 20;
                plot = 1;
              }
            }
            break;
          case XK_Down:
            /* volume down (also settable when muted) */
down:       if (g.vol_emphasized >= 0 && g.vol_emphasized < n) {
              if ((v[g.vol_emphasized].vol > -100 &&
                   v[g.vol_emphasized].vol > -60) ||
                  (v[g.vol_emphasized].vol < -100 &&
                   v[g.vol_emphasized].vol > -1000-60)) {
                v[g.vol_emphasized].vol--;
                volume_changed = 1;
                plot = 1;
              }
            }
            /* scroll view */
            else if (g.scrollbar_emphasized) {
              if (g.scroll_start > 0) {
                g.scroll_start -= 20;
                plot = 1;
              }
            }
            break;
        }
        break;
      }
      case ButtonPress:
        if (e.xbutton.button == 4) goto up;
        if (e.xbutton.button == 5) goto down;
        break;
      case MotionNotify:
        g.x = e.xmotion.x;
        g.y = e.xmotion.y;
        plot = 1;
        break;
      case EnterNotify:
        g.x = e.xcrossing.x;
        g.y = e.xcrossing.y;
        plot = 1;
        break;
      case LeaveNotify:
        g.x = -1;
        g.y = -1;
        plot = 1;
        break;
      }
    }
    if (volume_changed) {
      volume_changed = 0;
      write_int(fd, n);
      for (i = 0; i < n; i++) {
        write_int(fd, v[i].id);
        write_int(fd, v[i].vol);
      }
    }
    if (resize && (g.width != new_width || g.height != new_height)) {
      plot = 1;
      g.width = new_width;
      g.height = new_height;
      XftDrawDestroy(g.xft);
      XFreePixmap(g.d, g.p);
      g.p = XCreatePixmap(g.d, g.w, g.width, g.height,
          DefaultDepth(g.d, DefaultScreen(g.d)));
      XFillRectangle(g.d, g.p, g.colors[BG],
          0, 0, g.width, g.height);
      R((g.xft = XftDrawCreate(g.d, g.p,
          DefaultVisual(g.d, DefaultScreen(g.d)),
          DefaultColormap(g.d, DefaultScreen(g.d)))) == NULL);
    }
    resize = 0;
    timeout = -1;
    if (plot) {
      /* don't plot too much, max around 30Hz (1000/34 Hz to be precise) */
      uint64_t t = gettime();
      if (t - last_plot < 34) {
        timeout = t - last_plot;
      } else {
        last_plot = t;
        plot = 0;
        paint = 1;
        draw(&g, v, n);
      }
    }
    if (paint) {
      paint = 0;
      XCopyArea(g.d, g.p, g.w, g.colors[ACTIVE], 0,0, g.width, g.height, 0,0);
    }
    if (XPending(g.d)) goto x;
    R(poll(pfd, 2, timeout) == -1);
    R(pfd[0].revents & (POLLHUP | POLLERR));
    R(pfd[1].revents & (POLLHUP | POLLERR));
    if (pfd[1].revents & POLLIN) {
      plot = 1;
      free(v);
      n = get_int(fd);
      R((v = malloc(sizeof(vol) * n)) == NULL);
      for (i = 0; i < n; i++) {
        v[i].id = get_int(fd);
        v[i].vol = get_int(fd);
      }
      qsort(v, n, sizeof(vol), vol_cmp);
    }
  }

  return 0;
}
